﻿var widgets = require('@jupyter-widgets/base');
var _ = require('lodash');
var moment = require('moment');

const strRetrieving = 'Retrieving...'

// Import Font Awesome Icon Css
var iconCssId = 'iconCss';
if (!document.getElementById(iconCssId)) {
    var head = document.getElementsByTagName('head')[0];
    var link = document.createElement('link');
    link.id = iconCssId;
    link.rel = 'stylesheet';
    link.type = 'text/css';
    link.href = 'https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/css/font-awesome.min.css';
    link.media = 'all';
    head.appendChild(link);
}

var _template = {
    progressTemplate: function (value) {
        return `<div id="errorContainer" class="row" style="display:none; background-color: rgb(255,0,0); color:white; padding: 1px;"></div>
        <div class="progress">
        <div class="progress-bar progress-bar-striped active" role="progressbar" aria-valuenow="100" aria-valuemin="0" aria-valuemax="100" style="width:100%">
            ${value}
        </div>
        </div>`
    },
    statusColTemplate: function (value) {
        var retVal = value;
        switch (value) {
            case "Completed":
                retVal = "<span style='color:#2ca33e'>" + value + "</span>";
                break;
            case "Failed":
                retVal = "<span style='color:red'>" + value + "</span>";
                break;
            case "Scheduled":
            case "Running":
            case "Finalizing":
                retVal = "<span style='color:blue'>" + value + "</span>";
                break;
            case "Canceled":
                retVal = "<span style='color:#cfbd1d'>" + value + "</span>";
                break;
            default:
                break;
        }
        return retVal;
    },
    dateColTemplate: function (value) {
        var dt = moment(value);
        if (!dt.isValid()) {
            return "";
        }
        return dt.format("lll");
    },
    headTemplate: function (tooltip, title) {
        return $("<div>").prop("title", tooltip).text(title);
    },
    pipelineTemplate: function (value, item) {
        return $(`<a title="${item.run_properties}">${value}</a>`);
    },
    copyButtonTemplate: function (value) {
        return $("<i>").attr("class", "fa fa-clipboard").attr("run_id", value).prop("title", "Copy Run Id")
            .on("click", function (event) {
                var str = value;
                const el = document.createElement('textarea');
                el.value = str;
                document.body.appendChild(el);
                el.select();
                document.execCommand('copy');
                document.body.removeChild(el);
                event.stopPropagation();
            });
    },
    runIdTemplate: function (value) {
        return "<span run_id='" + value + "' class='run_id' style='overflow-wrap:break-word'>" + value + "</span>";
    }
};

var _commonModelProperties = {
    selected_run_id: '',
    run_properties: {},
    run_id: '',
    run_logs: '',
    run_metrics: [],
    status: '',
    run_arguments: '',
    child_runs: [],
    child_runs_metrics: [],
    run_process: '',
    compute_target_status: {}
}

var _statusColor = function (status) {
    switch (status.toLowerCase()) {
        case 'completed':
            return 'rgb(125,192,59)'
        case 'failed':
            return 'rgb(255,102,102)'
        case 'canceled':
            return 'rgb(255,255,0)'
        case 'starting':
            return 'rgb(236,157,21)'
        case 'running':
            return 'rgb(88,146,245)'
        default:
            return 'rgb(255,255,0)'
    }
}

var _renderLogs = function (element, logs) {
    if (!logs) return;
    element.text(logs);
    element[0].scrollTop = element[0].scrollHeight;
}

var append_pagination_dropdown = function (element, pageLength, childRunsLength, $menu) {
    if (pageLength >= childRunsLength) {
        $(element.find("#hdgrid")).append($menu);
    } else {
        $(element.find("#hdgrid .jsgrid-pager-container .jsgrid-pager ")).append($menu);
    }
    $menu.val(pageLength);
    $($menu[0]).off("change").on("change",
        function () {
            $(element.find("#hdgrid")).jsGrid("option", "pageSize", $(this).val());
            $(element.find("#hdgrid")).jsGrid("option", "pageIndex", 1);
        }
    );
}

var _renderJsGrid = function (element, jsGrid, data, fields, defaultSort, sortOrder, $menu) {
    var grid = element.find("#hdgrid");

    jsGrid.sortStrategies.timesorter = function (val1, val2) {
        var t1 = val1.split(':');
        var t2 = val2.split(':');
        var sec1 = (+t1[0]) * 60 * 60 + (+t1[1]) * 60 + (+t1[2]);
        var sec2 = (+t2[0]) * 60 * 60 + (+t2[1]) * 60 + (+t2[2]);
        return sec1 - sec2;
    };

    jsGrid.sortStrategies.number = function (val1, val2) {
        if(val1 === undefined)
            {return -1;}
        else if(val2 === undefined)
            {return 1;}
        else
            {return val1 - val2;}
    }

    // last page number and sort field during auto-refresh
    var pageIndex = grid.jsGrid("option", "pageIndex");
    var prevPageSize = _.toNumber(grid.jsGrid("option", "pageSize"));
    var sorting = grid.jsGrid("getSorting");

    grid.jsGrid({
        width: "100%",
        sorting: true,
        paging: true,
        pageSize: $menu && prevPageSize ? prevPageSize : 5,
        pageIndex: _.isInteger(pageIndex) ? pageIndex : 1,
        scrollOffset: 0,
        noDataContent: 'No runs scheduled yet',
        shrinkToFit: false,
        pagerFormat: "Pages: {first} {prev} {pages} {next} {last} &nbsp;&nbsp;",

        data: data,
        fields: fields,
        onRefreshed: function (grid) {
            if ($menu) {
                append_pagination_dropdown(element, this.pageSize, data.length, $menu);
            }
        }
    });

    if (sorting.field && fields.map(a=>a.name).indexOf(sorting.field) !== -1) {
        grid.jsGrid("sort", sorting);
    } else if (defaultSort && sortOrder) {
        grid.jsGrid("sort", {
            field: defaultSort,
            order: sortOrder
        });
    }
}

var _renderStatusChart = function (chartControl, element, child_runs) {
    if (!child_runs || child_runs.length == 0)
        return;

    var statusList = child_runs.reduce(function (prev, item) {
        if (item.status in prev) prev[item.status].count++;
        else prev[item.status] = {
            count: 1,
            color: _statusColor(item.status)
        };
        return prev;
    }, {});

    var chartContainerJQueryElement = element.find('#statusContainer');
    if (chartContainerJQueryElement.length > 0) {
        chartControl.horizontalBarChart(chartContainerJQueryElement[0], statusList, 'stack');
    }
}

var _renderBaseHTML = function (element, id) {
    if (element.find("#" + id).length < 1) {
        element.css('font-size', '14px');
        // popup container
        var container = element.find('.iziModal-content');
        if (container.length < 1) {
            container = element;
        }
        var baseHtml = `
        <style type="text/css">
        .EllipsisOverflow {
           text-overflow: ellipsis;
           overflow: hidden;
           white-space: nowrap;
        }
        </style>
        <div id="${id}" class="container-fluid">
            <div class="container-fluid">
                <div id="errorContainer" class="row" style="display:none; background-color: rgb(255,0,0); color:white; padding: 1px;"></div>
                <div class="row" style="display: flex;flex-wrap: wrap;">
                  <div class="col-sm-4" style="display: flex;flex-direction: column;">
                      <b>Run Properties</b>
                      <div class="table-responsive">
                          <table class="table table-striped" style="table-layout:fixed">
                              <tbody>
                                  <tr>
                                      <td>Status</td>
                                      <td><span id="jobStatus">${strRetrieving}</span></td>
                                  </tr>
                                  <tr id="computeTargetStatusContainer" style='display:none;'>
                                    <td>Compute Target Nodes</td>
                                    <td><div id="computeTargetStatus"></div></td>
                                  </tr>
                                  <tr>
                                      <td>Start Time</td>
                                      <td><span id="startTime">${strRetrieving}</span></td>
                                  </tr>
                                  <tr>
                                      <td>Duration</td>
                                      <td><span id="duration">${strRetrieving}</span></td>
                                  </tr>
                                  <tr>
                                    <td>Run Id</td>
                                    <td><span id="runId" style="word-wrap:break-word">${id}</span></td>
                                  </tr>
                              </tbody>
                          </table>
                      </div>
                  </div>
                  <div class="col-sm-8" style="display: flex;flex-direction: column;padding-left:5px;">
                      <div>
                        <div id="multiPrecisionRecallContainer" class="chart"></div>
                        <div id="rocContainer" class="chart"></div>
                        <div id='predictedVsTrueContainer' class='chart'></div>
                        <div id='confusionMatrixContainer' class='chart'></div>
                        <b>Output Logs</b>
                        <div id="logSelectorContainer" style="float: right; display:none;">
                            <span>Distributed run - Show log by process </span>
                            <select id="logSelector"></select>
                        </div>
                      </div>
                      <textarea id="txtDriverLog" class="form-control" style="min-width: 100%; height: 90%; font-size:inherit;" rows="10">Waiting for logs...</textarea>
                  </div>
                </div>
            </div>
        </div>`;
        container.append(baseHtml);
        return true;
    }
    return false;
}

var _populateBaseJobInfo = function (chartControl, element, job, compute_target_status) {
    var created_utc = new Date(job.created_utc);
    var run_duration = job.run_duration;
    var compute_target_status_container = $(element).find("#" + job.run_id).find("#computeTargetStatusContainer");
    $(element).find("#" + job.run_id).find("#jobStatus").text(job.status);
    // 1. Compute target can be reused in other run, and that means user can see this run is Completed and Compute Target is still runing other run
    //    It is better to hide it when this run is Completed
    // 2. For all types of node_state_counts are 0, pie chart can't be rendered at all and it is better to hde this segement
    if(compute_target_status && compute_target_status.node_state_counts && compute_target_status_container.length > 0 
        && job.status !== "Completed" && _.map(compute_target_status.node_state_counts).filter(e => e>0).length > 0){
        compute_target_status_container.show();
        var compute_target_status_element = compute_target_status_container.find('#computeTargetStatus')
        if(compute_target_status_element.length > 0){
            chartControl.pieChart(compute_target_status_element[0], compute_target_status.node_state_counts);
        }
    }
    else{
        compute_target_status_container.hide();
    }
    $(element).find("#" + job.run_id).find("#startTime").text(created_utc.toLocaleDateString() + " " + created_utc.toLocaleTimeString());
    $(element).find("#" + job.run_id).find("#duration").text(run_duration);
    $(element).find("#" + job.run_id).find("#runId").text(job.run_id ? job.run_id : "N/A");
    $(element).find("#" + job.run_id).find("#arguments").text(job.arguments ? job.arguments : "N/A");
    var logSelectorContainer = $(element).find("#" + job.run_id).find("#logSelectorContainer");

    if (!_.isEmpty(job.distributed_processes)) {
        var options = job.distributed_processes.map(e => '<option value="' + e + '">' + e + '</option>').join();
        logSelectorContainer.find('#logSelector').find('option').remove().end().append(options)
        logSelectorContainer.show();
    } else {
        logSelectorContainer.hide();
    }
}


var _renderError = function(elements, error) {
    if(error && elements.length > 0){
        $(elements[0]).text(error);
        $(elements[0]).show();
    }
    else{
        elements.hide();
    }
}

var _registerHandleClick = function () {
    return {
        'click table[class="jsgrid-table"] tr': '_handleClick',
        'change #logSelector': '_handleLogSelectorClick'
    };
}

var _handleClick = function (event) {
    event.preventDefault();
    var runIdElement = $(event.target).parents('tr').find('[run_id]');
    if (runIdElement.length > 0 && this.model && this.model.set) {
        this.model.set('selected_run_id', $(runIdElement).attr('run_id'));
        this.touch();
        // reset selected_run_id for the case that user clicks the same run which can't trigger change event
        // it is huge effort to connect back from close button of child single run widget to original widget, so just reset it right after
        this.model.set('selected_run_id', '');
        this.touch();
    }
}

var _handleLogSelectorClick = function (event) {
    event.preventDefault();
    if (this.model.set) {
        this.model.set('run_process', $(event.target).val());
        this.model.set('run_logs', 'Waiting for logs...');
        this.touch();
    }
}

var _popupElement = function (element, widget_settings) {
    if (_.get(widget_settings, 'display') !== 'popup') {
        return;
    }
    return setTimeout(function () {
        var $elem = $(element);
        var runId = $elem.find('#runId').text();
        $elem.iziModal({
            title: runId,
            width: $elem.width(),
            padding: 20,
            appendTo: 'body',
            closeOnEscape: true,
            closeButton: true
        });
        $elem.iziModal('open');
    });
}

var _getHashCode = function (s) {
    var hash = 0,
        i, chr;
    if (!s || s.length === 0) return hash;
    for (i = 0; i < s.length; i++) {
        chr = s.charCodeAt(i);
        hash = ((hash << 5) - hash) + chr;
        hash |= 0;
    }
    return hash;
};

class BackFillModel {
    constructor(value) {
        this.value = value;
    }
    get(property) {
        return this.value ? this.value[property] : null;
    }
}

// Include required CSS files
require('jsgrid/dist/jsgrid.min.css');
require('jsgrid/dist/jsgrid-theme.min.css');

var ShowRunDetailsModel = widgets.DOMWidgetModel.extend({
    defaults: _.extend(widgets.DOMWidgetModel.prototype.defaults(), Object.assign({
            _model_name: 'ShowRunDetailsModel',
            _view_name: 'ShowRunDetailsView',
            _model_module: 'azureml_train_widgets',
            _view_module: 'azureml_train_widgets',
            _model_module_version: "1.0.0",
            _view_module_version: "1.0.0",
        },
        _commonModelProperties
    ))
});

// Show Job Status View. Renders the widget model.
var ShowRunDetailsView = widgets.DOMWidgetView.extend({

    _handleClick: _handleClick,
    _handleLogSelectorClick: _handleLogSelectorClick,
    events: _registerHandleClick,
    renderer: null,

    render: function (value, el) {

        define("RenderHelper", ["jsgrid", "./chart_plotly"],
            function RenderHelper(jsgrid, ChartControl) {
                var renderer = function (element, widget_settings) {
                    this.element = element;
                    this.jsgrid = jsgrid;
                    this.ChartControl = ChartControl;
                    this.widget_settings = widget_settings
                }

                renderer.prototype.initRender = function () {
                    progress = this.element.find(".progress");
                    if (progress.length == 0) {
                        this.element.append(_template.progressTemplate('Loading widget...'));
                        _popupElement(this.element, this.widget_settings);
                    }
                }

                renderer.prototype.renderBaseHTML = function (run_id, workbench_uri) {
                    this.element.find(".progress").remove();
                    if (_renderBaseHTML(this.element, run_id)) {
                        var tableSelector = "table-striped";
                        var runTable = this.element.find("#" + run_id)
                        runTable.find("." + tableSelector).append(
                            `<tr>
                         <td>Arguments</td>
                         <td><div style="max-height:300px; overflow:auto;"><span id="arguments" style="word-wrap:break-word">N/A</span></div></td>
                      </tr>`);
                        runTable.append(`<div class='container-fluid'><div id='chartContainer' class='row'></div></div>
                                    <div id='statusContainer'></div>
                                    <div style='width:100%;display:inline-block' id='hdgrid'></div>
                                    <div id='metricContainer'></div>
                                    <div id='webLink'><a href='${workbench_uri}' target='_blank'>Click here to see the run in Azure portal</a></div>`);
                    }
                }

                renderer.prototype.populateRunProperties = function (run_properties, compute_target_status) {
                    _populateBaseJobInfo(this.ChartControl, this.element, run_properties, compute_target_status);
                }

                renderer.prototype.populateRunArguments = function (run_id, args) {
                    this.element.find("#" + run_id).find("#arguments").text(args ? args : "N/A");
                }

                renderer.prototype.renderRunMetrics = function (run_metrics, run_id) {
                    var element = this.element;
                    var plotChart = function (metricData, run_id) {
                        var chartId = _getHashCode(metricData.title);
                        if (element.find("#" + chartId).length < 1)
                            element.find("#" + run_id).find("#chartContainer").append("<div id='" + chartId + "' class='col-lg-4 col-md-4 col-sm-4 col-xs-6 chart' style='height:250px'></div>");
                        var chartElementJQueryObject = element.find("#" + chartId);
                        if (chartElementJQueryObject.length > 0) {
                            ChartControl.lineChart(chartElementJQueryObject[0], metricData);
                        }
                    }
                    var tableSelector = "table-striped";
                    var singleMetricSelector = "single-metric";
                    element.find("#" + run_id).find("." + singleMetricSelector).remove();
                    if (run_metrics && run_metrics.length > 0) {
                        run_metrics.forEach(function (metric) {
                            // single metric
                            if (metric && metric.series && metric.series[0] && metric.series[0].data && metric.series[0].data.length === 1) {
                                element.find("#" + metric.run_id).find("." + tableSelector).append(
                                    `<tr class=${singleMetricSelector}>
                                <td><div title='${metric.name}' class="EllipsisOverflow">${metric.name}</div></td>
                                <td><span>${metric.series[0].data[0]}</span></td>
                            </tr>`);
                            }
                            // series metric
                            else if (metric && metric.series && metric.series[0] && metric.series[0].data && metric.series[0].data.length > 1) {
                                var chartData = {
                                    title: metric.name,
                                    categories: metric.categories,
                                    series: metric.series
                                };
                                plotChart(chartData, metric.run_id);
                            }
                        });
                    }
                }

                renderer.prototype.renderLogs = function (run_id, run_logs) {
                    _renderLogs(this.element.find("#" + run_id).find("#txtDriverLog"), run_logs);
                }

                renderer.prototype.renderLeaderBoard = function (child_runs) {
                    var element = this.element;

                    if (child_runs && child_runs.length > 0) {
                        var fields = [{
                                name: "run_number",
                                title: "Run Number",
                                type: "number",
                                width: 110,
                                validate: "required"
                            },
                            {
                                name: "status",
                                title: "Status",
                                type: "text",
                                width: 100,
                                itemTemplate: _template.statusColTemplate
                            },
                            {
                                name: "start_time",
                                title: "Started",
                                type: "text",
                                width: 170,
                                itemTemplate: _template.dateColTemplate
                            },
                            {
                                name: "duration",
                                title: "",
                                type: "text",
                                width: 80,
                                headerTemplate: _template.headTemplate('Duration in hh:mm:ss', 'Duration'),
                                sorter: "timesorter"
                            },
                            {
                                name: "run_id",
                                title: "Run Id",
                                type: "text",
                                width: 400,
                                itemTemplate: _template.runIdTemplate
                            },
                        ];

                        _renderJsGrid(element, jsGrid, child_runs, fields);
                    }
                }

                renderer.prototype.renderChildrenMetricChart = function (child_runs_metrics) {
                    var element = this.element;
                    var chartData = child_runs_metrics;
                    if (!chartData || chartData.series == null)
                        return;

                    var chartContainerJQueryElement = element.find('#metricContainer');
                    chartContainerJQueryElement.find(".progress").remove();
                    if (chartContainerJQueryElement.length > 0) {
                        chartData.title = 'Run Metric : ' + chartData.primaryMetricName;
                        chartData.series = chartData.series[chartData.primaryMetricName]
                        ChartControl.lineChart(chartContainerJQueryElement[0], chartData, chartData.showLegend, null, 'category');
                    }
                }

                renderer.prototype.renderMultiPrecisionRecall = function (precision_recall_metrics_obj) {
                    var chartContainerJQueryElement = this.element.find('#multiPrecisionRecallContainer');
                    var precision_recall_metrics = _.get(precision_recall_metrics_obj, 'series[0].data[0]')
                    if (chartContainerJQueryElement.length > 0) {
                        ChartControl.multiPrecisionRecall(chartContainerJQueryElement[0], precision_recall_metrics);
                    }
                }

                renderer.prototype.renderROCCurve = function (roc_data_obj) {
                    var chartContainerJQueryElement = this.element.find('#rocContainer');
                    var roc_metrics = _.get(roc_data_obj, 'series[0].data[0]');
                    if (chartContainerJQueryElement.length > 0) {
                        ChartControl.ROCCurve(chartContainerJQueryElement[0], roc_metrics);
                    }
                }

                renderer.prototype.renderConfusionMatrix = function (confusion_matrices_obj, run_properties) {
                    if (!confusion_matrices_obj) {
                        return;
                    }

                    var confusion_matrices = _.get(confusion_matrices_obj, 'series[0].data[0]');
                    var c_labels = run_properties.properties.class_labels;
                    // example: if c_labels is "[1  2 3   4  ]", labels will be [1,2,3,4]
                    var labels = ((c_labels).slice(1, (c_labels).length - 1)).split(/\s+/);
                    var chartContainerJQueryElement = this.element.find('#confusionMatrixContainer');
                    if (chartContainerJQueryElement.length > 0) {
                        ChartControl.confusionMatrix(chartContainerJQueryElement[0], confusion_matrices, labels);
                    }
                }

                renderer.prototype.renderPredictedVsTrue = function (predicted_metrics_obj) {
                    if (!predicted_metrics_obj) {
                        return;
                    }

                    var chartContainerJQueryElement = this.element.find('#predictedVsTrueContainer');
                    var predicted_metrics = _.get(predicted_metrics_obj, 'series[0].data[0]')
                    if (chartContainerJQueryElement.length > 0) {
                        ChartControl.predictedVsTrueChart(chartContainerJQueryElement[0], predicted_metrics);
                    }
                }

            renderer.prototype.renderError = function (error) {
                _renderError(this.element.find("#errorContainer"), error)
			}

            return renderer;
    });

        if(this.model && this.model.on){
            this.model.on('change:run_id', this.run_id_changed, this);
            this.model.on('change:run_properties', this.run_properties_changed, this);
            this.model.on('change:run_logs', this.run_logs_changed, this);
            this.model.on('change:run_metrics', this.run_metrics_changed, this);
            this.model.on('change:child_runs', this.child_runs_changed, this);
            this.model.on('change:child_runs_metrics', this.child_runs_metrics_changed, this);
            this.model.on('change:compute_target_status', this.run_properties_changed, this);
            this.model.on('change:error', this.error_changed, this);
        }
        this.initialize_render(value, el);
    },

    child_runs_changed: function () {
        var child_runs = this.model.get('child_runs');
        this.renderer.renderLeaderBoard(child_runs);
    },

    child_runs_metrics_changed: function () {
        var child_runs_metrics = this.model.get('child_runs_metrics');
        this.renderer.renderChildrenMetricChart(child_runs_metrics);
    },

    run_id_changed: function () {
        var run_id = this.model.get('run_id');
        var workbench_uri = this.model.get('workbench_uri');
        this.renderer.renderBaseHTML(run_id, workbench_uri);
    },

    run_metrics_changed: function () {
        var run_properties = this.model.get('run_properties');
        var run_metrics = this.model.get('run_metrics');

        //Precision vs Recall chart
        const precision_recall_names = ["precision_recall"];
        const precision_recall_metrics =
            _.remove(run_metrics, function (metric) {
                return _.includes(precision_recall_names, metric.name);
            });
        if (!_.isEmpty(precision_recall_metrics)) {
            this.renderer.renderMultiPrecisionRecall(precision_recall_metrics[0]);
        }

        // Roc chart
        const roc_names = ["roc"];
        const roc_data = _.remove(run_metrics, function (metric) {
            return _.includes(roc_names, metric.name);
        });
        if (!_.isEmpty(roc_data)) {
            this.renderer.renderROCCurve(roc_data[0]);
        }

        // confusion matrix
        const confusion_matrices_names = ["confusion_matrices"];
        const confusion_matrices =
            _.remove(run_metrics, function (metric) {
                return _.includes(confusion_matrices_names, metric.name);
            });
        if (!_.isEmpty(confusion_matrices)) {
            this.renderer.renderConfusionMatrix(confusion_matrices[0], run_properties);
        }

        // predicted vs true chart
        const predictions_metrics_names = ["predictions"];
        const predictions_metrics =
            _.remove(run_metrics, function (metric) {
                return _.includes(predictions_metrics_names, metric.name);
            });
        if (!_.isEmpty(predictions_metrics)) {
            this.renderer.renderPredictedVsTrue(predictions_metrics[0], run_properties);
        }

        this.renderer.renderRunMetrics(run_metrics, run_properties.run_id);
    },

    run_logs_changed: function () {
        var run_id = this.model.get('run_id');
        var run_logs = this.model.get('run_logs');
        this.renderer.renderLogs(run_id, run_logs);
    },

    run_properties_changed: function () {
        var run_properties = this.model.get('run_properties');
        var compute_target_status = this.model.get('compute_target_status');
        this.renderer.populateRunProperties(run_properties, compute_target_status);
    },

    error_changed: function() {
        var error = this.model.get('error');
        this.renderer.renderError(error);
    },

    initialize_render: function (value, el) {
        var context = this;
        if (!context.el) {
            context.el = $(el);
        }
        if (!context.model) {
            context.model = new BackFillModel(value);
        }

        var widget_settings = context.model.get('widget_settings');
        var run_id = context.model.get('run_id');

        require(["jsgrid", "RenderHelper"], function (jsgrid, RenderHelper) {
            var initDashboard = function (context) {
                var element = $(context.el);

                context.renderer = new RenderHelper(element, widget_settings);
                context.renderer.initRender();

                // below is executed only when notebook is cached. Otherwise state will not have run_id set when initializing
                if (run_id) {
                    setTimeout(function () {
                        context.run_id_changed();
                        context.run_properties_changed();
                        context.run_logs_changed();
                        context.run_metrics_changed();
                        context.child_runs_changed();
                        context.child_runs_metrics_changed();
                        context.error_changed();
                    });
                }
            }

            initDashboard(context);
        });
    }
});

// HyperDrive Code Starts Here: ->

var ShowHyperDriveRunsModel = widgets.DOMWidgetModel.extend({
    defaults: _.extend(widgets.DOMWidgetModel.prototype.defaults(), Object.assign({
            _model_name: 'ShowHyperDriveRunsModel',
            _view_name: 'ShowHyperDriveRunsView',
            _model_module: 'azureml_train_widgets',
            _view_module: 'azureml_train_widgets',
            _model_module_version: "1.0.0",
            _view_module_version: "1.0.0",
        },
        _commonModelProperties))
});

// Show Job Status View. Renders the widget model.
var ShowHyperDriveRunsView = widgets.DOMWidgetView.extend({

    _handleClick: _handleClick,
    _handleLogSelectorClick: _handleLogSelectorClick,
    events: _registerHandleClick,
    renderer: null,

    render: function (value, el) {

        define("HyperDriveRenderHelper", ["jsgrid", "./chart_plotly"],
            function RenderHelper(jsgrid, ChartControl) {
                var renderer = function (element, widget_settings) {
                    this.element = element;
                    this.jsgrid = jsgrid;
                    this.ChartControl = ChartControl;
                    this.widget_settings = widget_settings;
                }

                var _getSortOrder = function (run_properties) {
                    var sortOrder;
                    try {
                        var config = JSON.parse(run_properties.properties.primary_metric_config);
                        if (config.goal === 'minimize') {
                            sortOrder = 'asc';
                        } else if (config.goal === 'maximize') {
                            sortOrder = 'desc';
                        }
                    } finally {
                        return sortOrder;
                    }
                }

                renderer.prototype.initRender = function () {
                    progress = this.element.find(".progress");
                    if (progress.length == 0)
                        this.element.append(_template.progressTemplate('Loading widget...'));
                }

                renderer.prototype.renderBaseHTML = function (run_id, workbench_uri) {
                    this.element.find(".progress").remove();

                    if (run_id && this.element.find("#metricContainer").length < 1) {
                        _renderBaseHTML(this.element, run_id);

                        var tableSelector = "table-striped";
                        var runTable = this.element.find("#" + run_id)
                        runTable.find("." + tableSelector).append(
                            `<tr>
                             <td>Max concurrent runs</td>
                             <td><span id="max_concurrent_jobs" style="word-wrap:break-word">${strRetrieving}</span></td>
                          </tr>
                          <tr>
                             <td>Max total runs</td>
                             <td><span id="max_total_jobs" style="word-wrap:break-word">${strRetrieving}</span></td>
                          </tr>`);

                        runTable.append(`<div id='statusContainer' class='chart'></div>
                          <div style='width:100%;display:inline-block' id='hdgrid'>${_template.progressTemplate('Waiting for runs...')}</div>* The best metric field is obtained from the min/max of primary metric achieved by a run<br/><br/>
                          <div id='metricContainer' class='chart'>${_template.progressTemplate('Waiting for metrics...')}</div>
                              <select class="switchChart" style="display:none">
                                  <option value="parallelCoordinates">Parallel Coordinates Chart</option>
                                  <option value="twoDimension">2D Scatter Chart</option>
                                  <option value="threeDimension">3D Scatter Chart</option>
                              </select>
                              <div class='parallelCoordinatesContainer chart' style="display:none"></div>
                              <div class='threeDimensionContainer chart' style="display:none"></div>
                              <div class='twoDimensionContainer chart' style="display:none"></div>
                          <div id='webLink'><a href='${workbench_uri}' target='_blank'>Click here to see the run in Azure portal</a></div>`);
                    }
                    return true;
                }

                renderer.prototype.populateRunProperties = function (run_properties, compute_target_status) {
                    _populateBaseJobInfo(this.ChartControl, this.element, run_properties, compute_target_status);

                    this.element.find("#" + run_properties.run_id).find("#max_concurrent_jobs").text(
                        (run_properties.tags && run_properties.tags.max_concurrent_jobs) ? run_properties.tags.max_concurrent_jobs : "N/A");
                    this.element.find("#" + run_properties.run_id).find("#max_total_jobs").text(
                        (run_properties.tags && run_properties.tags.max_total_jobs) ? run_properties.tags.max_total_jobs : "N/A");
                }

                renderer.prototype.renderLeaderBoard = function (child_runs, run_properties) {
                    var sortOrder = _getSortOrder(run_properties);
                    var allArguments = Object.keys(run_properties.hyper_parameters);

                    if (child_runs && child_runs.length > 0) {

                        var metricField, sortBy;
                        if(child_runs[0].best_metric && child_runs[0].metric){
                            metricField = {
                                name: "metric",
                                title: "Best Metric*",
                                type: "number",
                                width: 170
                            };
                            sortBy = "metric";
                        }
                        else{
                            metricField = {
                                name: "best_metric",
                                title: "Best Metric*",
                                type: "number",
                                width: 170
                            };
                            sortBy = "best_metric";
                        }

                        var fields = [{
                                name: "run_number",
                                title: "Run",
                                type: "number",
                                width: 70,
                                validate: "required"
                            },
                            metricField,
                            {
                                name: "status",
                                title: "Status",
                                type: "text",
                                width: 100,
                                itemTemplate: _template.statusColTemplate
                            },
                            {
                                name: "start_time",
                                title: "Started",
                                type: "text",
                                width: 170,
                                itemTemplate: _template.dateColTemplate
                            },
                            {
                                name: "duration",
                                title: "",
                                type: "text",
                                width: 80,
                                headerTemplate: _template.headTemplate('Queue and run duration in hh:mm:ss', 'Duration'),
                                sorter: "timesorter"
                            },
                            {
                                name: "run_id",
                                title: "Run Id",
                                type: "text",
                                width: 350,
                                itemTemplate: _template.runIdTemplate
                            },
                        ];

                        allArguments.forEach(key => {
                            fields.push({
                                name: "param_" + key,
                                title: key,
                                type: "number",
                                width: 200
                            });
                        });

                        _renderJsGrid(this.element, jsGrid, child_runs, fields, sortBy, sortOrder);
                    }
                }

                renderer.prototype.renderPrimaryMetricChart = function (child_runs, child_runs_metrics, run_properties) {
                    var element = this.element;
                    var chartData = child_runs_metrics;
                    if (!chartData || chartData.series == null)
                        return;

                    var chartContainerJQueryElement = element.find('#metricContainer');
                    chartContainerJQueryElement.find(".progress").remove();
                    if (chartContainerJQueryElement.length > 0) {
                        chartData.title = 'HyperDrive Run Primary Metric : ' + chartData.primaryMetricName;
                        chartData.series = chartData.series[chartData.primaryMetricName]
                        ChartControl.lineChart(chartContainerJQueryElement[0], chartData, chartData.showLegend, null, 'category', {
                            field: "metric",
                            order: _getSortOrder(run_properties)
                        });
                    }
                }

                renderer.prototype.renderParameterMetricChart = function (child_runs, child_runs_metrics, run_properties) {
                    var element = this.element;
                    if (!child_runs || !child_runs_metrics || child_runs_metrics.series == null)
                        return;

                    var switchElment = this.element.find('.switchChart');
                    var init = true;
                    switchElment.show();
                    switchElment.off().on('change', function () {
                        var chartData = {};
                        chartData.allArguments = Object.keys(run_properties.hyper_parameters);
                        if (init) {
                            switchElment.val((chartData.allArguments && chartData.allArguments.length >= 3) ?
                                'parallelCoordinates' : 'twoDimension');
                            init = false;
                        }
                        chartData.child_runs = child_runs;
                        chartData.metricName = child_runs_metrics.primaryMetricName;

                        var parallelCoordinatesContainerJQueryElement = element.find('.parallelCoordinatesContainer');
                        var threeDimensionContainerJQueryElement = element.find('.threeDimensionContainer');
                        var twoDimensionContainerJQueryElement = element.find('.twoDimensionContainer');

                        if (this.value !== 'threeDimension') {
                            ChartControl.purge(threeDimensionContainerJQueryElement);
                            threeDimensionContainerJQueryElement.hide();
                        }
                        if (this.value !== 'twoDimension') {
                            ChartControl.purge(twoDimensionContainerJQueryElement);
                            twoDimensionContainerJQueryElement.hide();
                        }
                        if (this.value !== 'parallelCoordinates') {
                            ChartControl.purge(parallelCoordinatesContainerJQueryElement);
                            parallelCoordinatesContainerJQueryElement.hide();
                        }

                        switch (this.value) {
                            case 'threeDimension':
                                threeDimensionContainerJQueryElement.show();
                                if (threeDimensionContainerJQueryElement.length > 0) {
                                    ChartControl.threeDimensionSurfaceChart(threeDimensionContainerJQueryElement[0], chartData, _getSortOrder(run_properties));
                                }
                                break;
                            case 'twoDimension':
                                twoDimensionContainerJQueryElement.show();
                                if (twoDimensionContainerJQueryElement.length > 0) {
                                    ChartControl.twoDimensionSurfaceChart(twoDimensionContainerJQueryElement[0], chartData, _getSortOrder(run_properties));
                                }
                                break;
                            case 'parallelCoordinates':
                            default:
                                parallelCoordinatesContainerJQueryElement.show();
                                if (parallelCoordinatesContainerJQueryElement.length > 0) {
                                    ChartControl.parallelCoordinatesChart(parallelCoordinatesContainerJQueryElement[0], chartData, _getSortOrder(run_properties));
                                }
                                break;
                        }
                    }).change();
                }

                renderer.prototype.renderStatusBar = function (child_runs) {
                    _renderStatusChart(this.ChartControl, this.element, child_runs);
                }

                renderer.prototype.renderLogs = function (run_logs) {
                    _renderLogs(this.element.find("#txtDriverLog"), run_logs);
                }

                renderer.prototype.renderError = function (error) {
                    _renderError(this.element.find("#errorContainer"), error)
                }

                return renderer;
            });

        if (this.model && this.model.on) {
            this.model.on('change:run_id', this.run_id_changed, this);
            this.model.on('change:run_properties', this.run_properties_changed, this);
            this.model.on('change:run_logs', this.run_logs_changed, this);
            this.model.on('change:child_runs', this.child_runs_changed, this);
            this.model.on('change:child_runs_metrics', this.child_runs_metrics_changed, this);
            this.model.on('change:compute_target_status', this.run_properties_changed, this);
            this.model.on('change:error', this.error_changed, this);
        }
        this.initialize_render(value, el);

    },

    child_runs_changed: function () {
        var child_runs = this.model.get('child_runs');
        var run_properties = this.model.get('run_properties');
        var child_runs_metrics = this.model.get('child_runs_metrics');
        this.renderer.renderLeaderBoard(child_runs, run_properties);
        this.renderer.renderStatusBar(child_runs);
        this.renderer.renderParameterMetricChart(child_runs, child_runs_metrics, run_properties);
    },

    child_runs_metrics_changed: function () {
        var child_runs_metrics = this.model.get('child_runs_metrics');
        var run_properties = this.model.get('run_properties');
        var child_runs = this.model.get('child_runs');
        this.renderer.renderPrimaryMetricChart(child_runs, child_runs_metrics, run_properties);
    },

    run_id_changed: function () {
        var run_id = this.model.get('run_id');
        var workbench_uri = this.model.get('workbench_uri');
        this.renderer.renderBaseHTML(run_id, workbench_uri);
    },

    run_logs_changed: function () {
        var run_logs = this.model.get('run_logs');
        this.renderer.renderLogs(run_logs);
    },

    run_properties_changed: function () {
        var run_properties = this.model.get('run_properties');
        var compute_target_status = this.model.get('compute_target_status');
        this.renderer.populateRunProperties(run_properties, compute_target_status);
    },

    error_changed: function() {
        var error = this.model.get('error');
        this.renderer.renderError(error);
    },

    initialize_render: function (value, el) {
        var context = this;
        if (!context.el) {
            context.el = $(el);
        }
        if (!context.model) {
            context.model = new BackFillModel(value);
        }
        var widget_settings = context.model.get('widget_settings');
        var run_id = context.model.get('run_id');

        require(["jsgrid", "HyperDriveRenderHelper"], function (jsgrid, HyperDriveRenderHelper) {
            var initDashboard = function (context) {
                var element = $(context.el);

                context.renderer = new HyperDriveRenderHelper(element, widget_settings);
                context.renderer.initRender();

                // below is executed only when notebook is cached. Otherwise state will not have run_id set when initializing
                if (run_id) {
                    setTimeout(function () {
                        context.run_id_changed();
                        context.run_properties_changed();
                        context.run_logs_changed();
                        context.child_runs_changed();
                        context.child_runs_metrics_changed();
                        context.error_changed();
                    });
                }
            }

            initDashboard(context);
        });
    }
});
// HyperDrive Code Ends Here: <-

var ShowAutoMLRunsModel = widgets.DOMWidgetModel.extend({
    defaults: _.extend(widgets.DOMWidgetModel.prototype.defaults(), Object.assign({
            _model_name: 'ShowAutoMLRunsModel',
            _view_name: 'ShowAutoMLRunsView',
            _model_module: 'azureml_train_widgets',
            _view_module: 'azureml_train_widgets',
            _model_module_version: "1.0.0",
            _view_module_version: "1.0.0",
        },
        _commonModelProperties))
});

var ShowAutoMLRunsView = widgets.DOMWidgetView.extend({

    _handleClick: _handleClick,
    _handleLogSelectorClick: _handleLogSelectorClick,
    events: _registerHandleClick,
    renderer: null,

    render: function (value, el) {

        define("AutoMLRenderHelper", ["jsgrid", "./chart_plotly"],
            function RenderHelper(jsgrid, ChartControl) {
                var renderer = function (element, widget_settings) {
                    this.element = element;
                    this.jsgrid = jsgrid;
                    this.ChartControl = ChartControl;
                    this.widget_settings = widget_settings;
                }

                var _getSortOrder = function (child_runs, run_properties) {
                    var sortOrder;
                    try {
                        var goal = child_runs[0].goal;
                        var primaryMetric = run_properties.properties.primary_metric;
                        if (goal === (primaryMetric + '_max')) {
                            sortOrder = 'desc';
                        } else if (goal === (primaryMetric + '_min')) {
                            sortOrder = 'asc';
                        }
                    } finally {
                        return sortOrder;
                    }
                }

                renderer.prototype.initRender = function () {
                    progress = this.element.find(".progress");
                    if (progress.length == 0) {
                        this.element.append(_template.progressTemplate('Loading widget...'));
                    }
                }
                renderer.prototype.renderBaseHTML = function (run_id, workbench_uri) {
                    this.element.find(".progress").remove();

                    if (run_id && this.element.find("#metricContainer").length < 1) {
                        this.element.append(`<div id='parentRunContainer'></div>
                        <div id='statusContainer' class='chart'></div>
                        <div style='width:100%;display:inline-block' id='hdgrid'>${_template.progressTemplate('Waiting for runs...')}</div>
                        <div id='dropDownContainer'></div>
                        <div id='metricContainer' class='chart'>${_template.progressTemplate('Waiting for metrics...')}</div>
                        <div id='webLink'><a href='${workbench_uri}' target='_blank'>Click here to see the run in Azure portal</a></div>`);
                    }
                    return true;
                }

                renderer.prototype.renderHeatmapBar = function (child_runs, child_runs_metrics) {
                    if (!child_runs || child_runs.length == 0) {
                        return;
                    }

                    var chartData = {
                        child_runs: child_runs,
                        colors: {
                            lightgreen: '#a8e063',
                            green: '#56ab2f',
                            gray: '#eeeeee',
                            red: '#ea5455',
                            blue: '#5b86e5'
                        }
                    };

                    var chartContainerJQueryElement = this.element.find('#statusContainer');
                    if (chartContainerJQueryElement.length > 0) {
                        var primaryMetricName = _.get(child_runs_metrics, 'primaryMetricName');
                        var series = _.get(child_runs_metrics, ['series', primaryMetricName]);
                        var categories = _.get(child_runs_metrics, 'categories');
                        chartData.child_runs_metrics = {
                            series: series,
                            categories: categories
                        }
                        ChartControl.heatmapBarChart(chartContainerJQueryElement[0], chartData);
                    }
                }
                renderer.prototype.renderParentRunStatus = function (run_properties) {
                    if (!run_properties || !run_properties['run_id'] || !run_properties['status']) {
                        return;
                    }

                    var run_id = run_properties['run_id'];
                    var parentRunContainer = this.element.find('#parentRunContainer');
                    if (parentRunContainer.find('#parentRunID').length === 0) {
                        parentRunContainer.append(`<div id="parentRunID" style="fontWeight:bold; font-size:18px;">${run_id}:</div>
                        <div id="parentStatusContainer"></div>`);
                    }
                    var status = run_properties['status'];
                    parentRunContainer.find('#parentStatusContainer').html(`Status: ${_template.statusColTemplate(status)}`);

                    var errors = _.get(run_properties, 'properties.errors');
                    if (status === "Failed" && errors && this.element.find('#hdgrid')) {
                        this.element.find('.progress').remove();
                        parentRunContainer.append(`<div id="parentRunErrorMsg" style="color:red">${errors}</div>`);
                    }
                }

                renderer.prototype.renderLeaderBoard = function (child_runs, run_properties) {
                    var element = this.element;
                    var sortOrder = _getSortOrder(child_runs, run_properties);

                    function checkStartDate(dt) {
                        return dt.start_time && dt.start_time != 'None'
                    }

                    var stVisible = child_runs && child_runs.some(checkStartDate);

                    var fields = [{
                            name: "iteration",
                            title: "Iteration",
                            type: "number",
                            width: 70,
                            validate: "required"
                        },
                        {
                            name: "run_name",
                            title: "Pipeline",
                            type: "text",
                            width: 300,
                            itemTemplate: _template.pipelineTemplate
                        },
                        {
                            name: "primary_metric",
                            title: "Iteration metric",
                            type: "number",
                            width: 130
                        },
                        {
                            name: "best_metric",
                            title: "",
                            type: "number",
                            width: 100,
                            headerTemplate: _template.headTemplate('Best metric value reached at the time of the iteration', 'Best metric')
                        },
                        {
                            name: "status",
                            title: "Status",
                            type: "text",
                            width: 100,
                            itemTemplate: _template.statusColTemplate
                        },
                        {
                            name: "duration",
                            title: "",
                            type: "text",
                            width: 80,
                            headerTemplate: _template.headTemplate('Duration in hh:mm:ss', 'Duration'),
                            sorter: "timesorter"
                        },
                        {
                            name: "start_time",
                            title: "Started",
                            type: "text",
                            width: 170,
                            itemTemplate: _template.dateColTemplate,
                            visible: stVisible
                        },
                        {
                            name: "run_id",
                            title: "Run Id",
                            type: "text",
                            align: "center",
                            width: 60,
                            itemTemplate: _template.copyButtonTemplate
                        },
                    ];

                    var default_val = 5;

                    var $menu = $(`<select id ='pageDropDown'>
                        <option value="1">1</option>
                        <option value="5">5</option>
                        <option value="10">10</option>
                        <option value="20">20</option>
                        <option value="50">50</option>
                    <\select>
                    <span style="white-space:nowrap; font-size:14px;" id='paginationText'> per page</span>`);
                    if ($(element.find("#pageDropDown")).length < 1) {
                        $menu.val(default_val);
                        var grid = element.find("#hdgrid");
                        grid.jsGrid({
                            pageSize: default_val
                        });
                    }
                    _renderJsGrid(element, jsGrid, child_runs, fields, "primary_metric", sortOrder, $menu);
                }

                renderer.prototype.renderPrimaryMetricChart = function (child_runs_metrics) {
                    var chartData = child_runs_metrics;
                    if (!chartData || chartData.series == null)
                        return;

                    var chartContainerJQueryElement = this.element.find('#metricContainer');
                    chartContainerJQueryElement.find(".progress").remove();
                    if (this.element.find('#pm_dropdown').length > 0) {
                        chartData.primaryMetricName = $( "#pm_dropdown option:selected" ).text();
                    }
                    if (chartContainerJQueryElement.length > 0) {
                        chartData.title = 'AutoML Run Primary Metric : ' + chartData.primaryMetricName;
                        chartData.series.categories = chartData.categories;
                        chartData.series = chartData.series;
                        ChartControl.changeableLineChart(chartContainerJQueryElement[0], this.element.find('#dropDownContainer'), chartData, false);
                    }
                }

                renderer.prototype.renderError = function (error) {
                    _renderError(this.element.find("#errorContainer"), error)
                }

                return renderer;
            });
        if (this.model && this.model.on) {
            this.model.on('change:run_id', this.run_id_changed, this);
            this.model.on('change:run_properties', this.run_properties_changed, this);
            this.model.on('change:child_runs', this.child_runs_changed, this);
            this.model.on('change:child_runs_metrics', this.child_runs_metrics_changed, this);
            this.model.on('change:compute_target_status', this.run_properties_changed, this);
            this.model.on('change:error', this.error_changed, this);
        }
        this.initialize_render(value, el);
    },

    child_runs_changed: function () {
        var child_runs = this.model.get('child_runs');
        var run_properties = this.model.get('run_properties');
        var child_runs_metrics = this.model.get('child_runs_metrics');
        this.renderer.renderLeaderBoard(child_runs, run_properties);
        this.renderer.renderHeatmapBar(child_runs, child_runs_metrics);
    },

    child_runs_metrics_changed: function () {
        var child_runs_metrics = this.model.get('child_runs_metrics');
        var child_runs = this.model.get('child_runs');
        this.renderer.renderHeatmapBar(child_runs, child_runs_metrics);
        this.renderer.renderPrimaryMetricChart(child_runs_metrics);
    },

    run_id_changed: function () {
        var run_id = this.model.get('run_id');
        var workbench_uri = this.model.get('workbench_uri');
        this.renderer.renderBaseHTML(run_id, workbench_uri);

    },

    run_properties_changed: function () {
        var run_properties = this.model.get('run_properties');
        this.renderer.renderParentRunStatus(run_properties);
    },

    error_changed: function() {
        var error = this.model.get('error');
        this.renderer.renderError(error);
    },

    initialize_render: function (value, el) {
        var context = this;
        if (!context.el) {
            context.el = $(el);
        }
        if (!context.model) {
            context.model = new BackFillModel(value);
        }
        var widget_settings = context.model.get('widget_settings');
        var run_id = context.model.get('run_id');

        require(["jsgrid", "AutoMLRenderHelper"], function (jsgrid, AutoMLRenderHelper) {
            var initDashboard = function (context) {
                var element = $(context.el);

                context.renderer = new AutoMLRenderHelper(element, widget_settings);
                context.renderer.initRender();

                // below is executed only when notebook is cached. Otherwise state will not have run_id set when initializing
                if (run_id) {
                    setTimeout(function () {
                        context.run_id_changed();
                        context.run_properties_changed();
                        context.child_runs_changed();
                        context.child_runs_metrics_changed();
                        context.error_changed();
                    });
                }
            }

            initDashboard(context);
        });

    }
});

module.exports = {
    ShowRunDetailsModel: ShowRunDetailsModel,
    ShowRunDetailsView: ShowRunDetailsView,
    ShowHyperDriveRunsModel: ShowHyperDriveRunsModel,
    ShowHyperDriveRunsView: ShowHyperDriveRunsView,
    ShowAutoMLRunsModel: ShowAutoMLRunsModel,
    ShowAutoMLRunsView: ShowAutoMLRunsView
};